unit ElMTree;

interface

uses classes, ElList;

type TElMTree = class;

     TElMTreeItem = ^RElMTreeItem;
     RElMTreeItem = record
        Parent: TElMTreeItem;
        Data  : pointer;
        List  : TElList;
     end;

     TIterProc = procedure (Item:TElMTreeItem; Index: integer; var ContinueIterate:boolean;
                            IterateData:pointer);


     TElMTree = class
       private
         FRoot : TElMTreeItem;
         FCount : integer;
       private
         function GetItem(index : integer) : TElMTreeItem;
       public
         constructor Create;
         destructor Destroy; override;
         function AddItem(Parent : TElMTreeItem; Value : pointer) : TElMTreeItem;
         function InsertItem(Parent : TElMTreeItem; Index : integer; Value : pointer) : TElMTreeItem;
         procedure DeleteItem(Item : TElMTreeItem);
         procedure MoveTo(Item, NewParent : TElMTreeItem);
         procedure Clear;
         function GetIndex(Item : TElMTreeItem) : Integer;
         function GetAbsIndex(Item : TElMTreeItem) : Integer;
         procedure Iterate(IterateProc:TIterProc; IterateData:pointer);

         property Count : Integer read FCount;  { Public }
         property Item[index : integer] : TElMTreeItem read GetItem;  { Public }
     end;

implementation

procedure TElMTree.Iterate(IterateProc:TIterProc; IterateData:pointer);
var j:integer;
    DoContinue: boolean;

  procedure IntIterate(Item:TElMTreeItem);
  var i:integer;
  begin
    inc(j);
    if j>=0 then IterateProc(Item, j, DoContinue, IterateData);
    if not (DoContinue) then exit;
    for i:=0 to Item^.List.Count-1 do
      begin
        IntIterate(TElMTreeItem(Item.List[i]));
        if not (DoContinue) then exit;
      end;
  end;

begin
  j:=-2;
  DoContinue:=true;
  IntIterate(FRoot);
end;

function TElMTree.GetItem(index : integer) : TElMTreeItem;
type PGIRec = ^TGIRec;
     TGIRec = record
       j:integer;
       TSI:TElMTreeItem;
     end;
var GIRec : TGIRec;

  procedure IntGetItem(Item:TElMTreeItem; Index: integer; var ContinueIterate:boolean;
                       IterateData:pointer);
  begin
    if Index=PGIRec(IterateData)^.j then
    begin
      PGIRec(IterateData)^.TSI:=Item;
      ContinueIterate:=false;
    end;
  end;

begin
  if (index<0) or (index>=FCount) then
  begin
    result:=nil;
    exit;
  end;
  GIRec.TSI:=nil;
  GIRec.j:=index;
  Iterate(@IntGetItem, @GIRec);
  result:=GIRec.TSI;
end;

function TElMTree.AddItem(Parent : TElMTreeItem; Value : pointer) : TElMTreeItem;  { public }
var TI : TElMTreeItem;
begin
  New(TI);
  TI^.List:=TElList.Create;
  TI^.Data:=Value;
  if Parent = nil then FRoot^.List.Add(TI) else Parent^.List.Add(TI);
  if Parent = nil then TI^.Parent:=FRoot else TI^.Parent:=Parent;
  result:=ti;
  inc(FCount);
end;  { AddItem }

function TElMTree.InsertItem(Parent : TElMTreeItem; Index : integer; Value : pointer) : TElMTreeItem;  { public }
var TI : TElMTreeItem;
begin
  New(TI);
  TI^.List:=TElList.Create;
  TI^.Data:=Value;
  if Parent = nil then
  begin
    FRoot^.List.Insert(index, TI);
    TI^.Parent:=FRoot;
  end
  else
  begin
    Parent^.List.Insert(Index, TI);
    TI^.Parent:=Parent;
  end;
  result:=ti;
  inc(FCount);
end;  { InsertItem }

procedure TElMTree.DeleteItem(Item : TElMTreeItem);  { public }
begin
  if (Item = nil) then exit;
  while Item^.List.Count>0 do DeleteItem(Item^.List[0]);
  Item^.List.Free;
  Item^.Parent^.List.Remove(Item);
  Dispose(Item);
  dec(FCount);  
end;  { DeleteItem }

procedure TElMTree.Clear;  { public }
begin
  while FRoot^.List.Count>0 do DeleteItem(FRoot^.List[0]);
end;  { Clear }

procedure TElMTree.MoveTo(Item, NewParent : TElMTreeItem);  { public }
begin
  if Item=nil then exit;
  Item^.Parent.List.Remove(Item);
  NewParent^.List.Add(Item);
  Item^.Parent:=NewParent;
end;  { MoveTo }

function TElMTree.GetIndex(Item : TElMTreeItem) : Integer;  { public }
begin
  result:=Item^.Parent^.List.IndexOf(Item);
end;  { GetIndex }

function TElMTree.GetAbsIndex(Item : TElMTreeItem) : Integer;  { public }
type PGIRec = ^TGIRec;
     TGIRec = record
       j:integer;
       TSI:TElMTreeItem;
     end;
var GIRec : TGIRec;

  procedure IntGetIndex(Item:TElMTreeItem; Index: integer; var ContinueIterate:boolean;
                                          IterateData:pointer);
  begin
    if PGIRec(IterateData)^.TSI=Item then
    begin
      PGIRec(IterateData)^.j:=index;
      ContinueIterate:=false;
    end;
  end;

begin
  if Item=nil then begin result:=-1; exit; end;
  GIRec.j:=-1;
  GIRec.TSI:=Item;
  Iterate(@IntGetIndex, @GIRec);
  result:=GIRec.j;
end;

constructor TElMTree.Create;  { public }
begin
  inherited;
  New(FRoot);
  FRoot^.List:=TElList.Create;
end;  { Create }

destructor TElMTree.Destroy;  { public }
begin
  FRoot^.List.Free;
  Dispose(FRoot);
  inherited;
end;  { Destroy }

end.
 
